############################################################################
# 
# Copyright (c) 2001 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
############################################################################

import re, time, string, sys, marshal, cPickle, base64
import Globals
from OFS.SimpleItem import Item
from Acquisition import Implicit, Explicit, aq_base
from Persistence import Persistent
from AccessControl.Owned import Owned
from AccessControl.Role import RoleManager
from App.Management import Tabs
from zLOG import LOG, WARNING, BLATHER
from AccessControl import ClassSecurityInfo
from types import StringType
from ZPublisher.BeforeTraverse import registerBeforeTraverse, \
    unregisterBeforeTraverse

from Products.Sessions.BrowserIdManager import isAWellFormedBrowserId, \
     getNewBrowserId, BROWSERID_MANAGER_NAME
from Products.Sessions.SessionPermissions import *
from Products.Sessions import SessionInterfaces
from Products.Sessions.SessionDataManager import SessionDataManagerErr, \
    SessionDataManagerTraverser
from Products.Sessions.common import DEBUG

bad_path_chars_in=re.compile('[^a-zA-Z0-9-_~\,\. \/]').search

constructSessionDataManagerForm = Globals.DTMLFile('dtml/addDataManager',
    globals())

ADD_SESSION_DATAMANAGER_PERM="Add Session Data Manager"

def constructSessionDataManager(self, id, title='', path=None,
                                tableName='', requestName=None,
                                REQUEST=None):
    """ """
    ob = ZSQLSessionDataManager(id, path, title, requestName)
    self._setObject(id, ob)
    if not tableName.strip():
        tableName = 'sessions_' + id
    self._getOb(id).setTableName(tableName)
    if REQUEST is not None:
        return self.manage_main(self, REQUEST, update_menu=1)

class ZSQLSessionDataManager(Item, Implicit, Persistent, RoleManager,
        Owned, Tabs):
    """ The ZSQL Session Data Manager implementation """

    meta_type = 'Session Data Manager (ZSQL)'

    manage_options=(
        {'label': 'Settings',
         'action':'manage_sessiondatamgr',
         },
        {'label': 'Security',
         'action':'manage_access',
         },
        {'label': 'Ownership',
         'action':'manage_owner'
         },
        )

    security = ClassSecurityInfo()
    security.declareObjectPublic()

    ok = {'meta_type':1, 'id':1, 'title': 1, 'icon':1,
          'bobobase_modification_time':1, 'title_or_id':1 }
    security.setDefaultAccess(ok)
    security.setPermissionDefault(CHANGE_DATAMGR_PERM, ['Manager'])
    security.setPermissionDefault(MGMT_SCREEN_PERM, ['Manager'])
    security.setPermissionDefault(ACCESS_CONTENTS_PERM,['Manager','Anonymous'])
    security.setPermissionDefault(ARBITRARY_SESSIONDATA_PERM,['Manager'])
    security.setPermissionDefault(ACCESS_SESSIONDATA_PERM,
                                  ['Manager','Anonymous'])

    icon='misc_/CoreSessionTracking/datamgr.gif'

    __implements__ = (SessionInterfaces.SessionDataManagerInterface, )

    manage_sessiondatamgr = Globals.DTMLFile('dtml/manageDataManager',
        globals())

    # INTERFACE METHODS FOLLOW

    security.declareProtected(ACCESS_SESSIONDATA_PERM, 'getSessionData')
    def getSessionData(self, create=1):
        """ """
        key = self.getBrowserIdManager().getBrowserId(create=create)
        if key is not None:
            return self._getSessionDataObject(key)

    security.declareProtected(ACCESS_SESSIONDATA_PERM, 'hasSessionData')
    def hasSessionData(self):
        """ """
        key = self.getBrowserIdManager().getBrowserId(create=0)
        if key is None:
            return 0
        return self._hasSessionDataObject(key)

    security.declareProtected(ARBITRARY_SESSIONDATA_PERM,'getSessionDataByKey')
    def getSessionDataByKey(self, key):
        return self._getSessionDataObjectByKey(key)
    
    security.declareProtected(ACCESS_CONTENTS_PERM, 'getBrowserIdManager')
    def getBrowserIdManager(self):
        """ """
        mgr = getattr(self, BROWSERID_MANAGER_NAME, None)
        if mgr is None:
            raise SessionDataManagerErr,(
                'No browser id manager named %s could be found.' %
                BROWSERID_MANAGER_NAME
                )
        return mgr

    # END INTERFACE METHODS

    tableName = None   
    def __init__(self, id, path=None, title='', requestName=None):
        self.id = id
        self.setConnectionPath(path)
        self.setTitle(title)
        self._requestSessionName = requestName

    security.declareProtected(CHANGE_DATAMGR_PERM, 'manage_changeSDM')
    def manage_changeSDM(self, title, path=None, requestName=None,
                         tableName=None, REQUEST=None):
        """ """
        self.setConnectionPath(path)
        self.setTitle(title)
        self.setTableName(tableName)
        if requestName:
            if requestName != self._requestSessionName:
                self.updateTraversalData(requestName)
        else:
            self.updateTraversalData(None)
        if REQUEST is not None:
            return self.manage_sessiondatamgr(
                self, REQUEST, manage_tabs_message = 'Changes saved.'
                )

    security.declareProtected(CHANGE_DATAMGR_PERM, 'manage_clearOldSessions')
    def manage_clearOldSessions(self, age, REQUEST=None):
        """ """
        c = self._getSessionDataConnection()
        date = time.time() - (age * 24 * 60 * 60)
        date = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(date))
        sql = "select count(*) from %s where lastmod < '%s'"
        r = c.query(sql%(self.tableName, date), max_rows=1)
        num = r[1][0][0]

        sql = "delete from %s where lastmod < '%s'"
        r = c.query(sql%(self.tableName, date))
        if REQUEST is not None:
            return self.manage_sessiondatamgr(self, REQUEST,
                manage_tabs_message='%s sessions deleted.'%num)


    security.declareProtected(CHANGE_DATAMGR_PERM, 'setTitle')
    def setTitle(self, title):
        """ """
        if not title: self.title = ''
        else: self.title = str(title)

    security.declareProtected(CHANGE_DATAMGR_PERM, 'setConnectionPath')
    def setConnectionPath(self, path=None):
        """ """
        if not path:
            self.obpath = None # undefined state
        elif type(path) is type(''):
            if bad_path_chars_in(path):
                raise SessionDataManagerErr, (
                    'Connection path contains characters invalid in a Zope '
                    'object path'
                    )
            self.obpath = string.split(path, '/')
        elif type(path) in (type([]), type(())):
            self.obpath = list(path) # sequence
        else:
            raise SessionDataManagerErr, ('Bad path value %s' % path)

    security.declareProtected(CHANGE_DATAMGR_PERM, 'setTableName')
    def setTableName(self, tableName):
        """ """
        if tableName == self.tableName:
            return

        c = self._getSessionDataConnection()

        # drop existing table
        if self.tableName is not None:
            sql = "drop table %s"
            c.query(sql%self.tableName)

        # create session table
        self.tableName = tableName
        sql = '''create table %s (
            bid TEXT, key TEXT, value TEXT, lastmod DATE,
            primary key (bid, key)
        )'''
        c.query(sql%self.tableName)

    security.declareProtected(CHANGE_DATAMGR_PERM, 'getTableName')
    def getTableName(self):
        """ """
        return self.tableName

    security.declareProtected(MGMT_SCREEN_PERM, 'getConnectionPath')
    def getConnectionPath(self):
        """ """
        if self.obpath is not None:
            return string.join(self.obpath, '/')
        return '' # blank string represents undefined state
    
    def _hasSessionDataObject(self, key):
        """ """
        c = self._getSessionDataConnection()
        c.query('select count(*) from %s where key=%%s'%self.tableName,
            (key, ))
        return c.fetchone()[0]

    def _getSessionDataObject(self, key):
        """ returns new or existing session data object """
        c = self._getSessionDataConnection()
        return SessionDataProxy(c, key).__of__(self)

    def _getSessionDataObjectByKey(self, key):
        """ returns new or existing session data object """
        if not self._hasSessionDataObject(key):
            return None
        return self._getSessionDataObject(key)

    def _getSessionDataConnection(self):
        """ Do not cache the results of this call.  Doing so breaks the
        transactions for mounted storages. """
        if self.obpath is None:
            err = 'Session data container is unspecified in %s' % self.getId()
            LOG('Session Tracking', WARNING, err)
            raise SessionIdManagerErr, err
        try:
            return self.restrictedTraverse(self.obpath)()
        except:
            raise
            raise SessionDataManagerErr, (
                "External session data container '%s' not found." %
                string.join(self.obpath,'/')
                )

    security.declareProtected(MGMT_SCREEN_PERM, 'getRequestName')
    def getRequestName(self):
        """ """
        return self._requestSessionName or ''

    def manage_afterAdd(self, item, container):
        """ Add our traversal hook """
        self.updateTraversalData(self._requestSessionName)

    def manage_beforeDelete(self, item, container):
        """ Clean up on delete """
        self.updateTraversalData(None)

    def updateTraversalData(self, requestSessionName=None):
        # Note this cant be called directly at add -- manage_afterAdd will
        # work though.
        parent = self.aq_inner.aq_parent

        if getattr(self,'_hasTraversalHook', None):
            unregisterBeforeTraverse(parent, 'SessionDataManager')
            del self._hasTraversalHook
            self._requestSessionName = None

        if requestSessionName:
            hook = SessionDataManagerTraverser(requestSessionName, self.id)
            registerBeforeTraverse(parent, hook, 'SessionDataManager', 50)
            self._hasTraversalHook = 1
            self._requestSessionName = requestSessionName

Globals.InitializeClass(ZSQLSessionDataManager)


class SessionDataProxy(Implicit):
    def __init__(self, connection, browser_id):
        self.connection = connection
        self.browser_id = browser_id

    def has_key(self, key):
        qkey = base64.encodestring(marshal.dumps(key)).strip()
        sql = "select count(*) from %s where bid='%s' and key='%s'"
        r = self.connection.query(sql%(self.tableName, self.browser_id,
            qkey), max_rows=1)
        return r[1][0][0]

    def __getitem__(self, key):
        qkey = base64.encodestring(marshal.dumps(key)).strip()
        sql = "select value from %s where bid='%s' and key='%s'"
        r = self.connection.query(sql%(self.tableName, self.browser_id,
            qkey), max_rows=1)
        if not r or not r[1]:
            raise KeyError, key
        value = cPickle.loads(base64.decodestring(r[1][0][0]))
        return value

    _marker = []
    def get(self, key, default=_marker):
        if self.has_key(key):
            return self[key]
        elif default is self._marker:
            return None
        return default

    def __setitem__(self, key, value):
        qkey = base64.encodestring(marshal.dumps(key)).strip()
        qvalue = base64.encodestring(cPickle.dumps(value)).strip()
        lastmod = time.strftime('%Y-%m-%d %H:%M:%S')
        if self.has_key(key):
            sql = """update %s set value='%s', lastmod='%s'
            where bid='%s' and key='%s'"""
            self.connection.query(sql%(self.tableName, qvalue,
                self.browser_id, qkey))
        else:
            sql = """insert into %s (bid, key, value, lastmod) values 
                ('%s', '%s', '%s', '%s')"""
            self.connection.query(sql%(self.tableName, self.browser_id,
                qkey, qvalue, lastmod))
    set = __setitem__

    def __delitem__(self, key):
        qkey = base64.encodestring(marshal.dumps(key)).strip()
        sql = "delete from %s where bid='%s' and key='%s'"
        self.connection.query(sql%(self.tableName, self.browser_id, qkey))

